/* Jon Arney, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#endif

#include <assert.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <stdarg.h>

#include "gnut_lib.h"
#include "gnut_connection.h"
#include "gnut_net.h"
#include "host.h"
#include "gnut_transfer.h"
#include "query.h"
#include "conf.h"
#include "share.h"
#include "gnut_threads.h"
#include "gnut_http.h"
#include "Gnut_Hash.h"

void fre_hrt(http_request_t **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

void fre_hct(http_cookie_t **x, int bugnum)
{
  yfre((void **) x, bugnum);
}

http_request_t *http_request_new(int sock)
{
  http_request_t *req;

  gd_s(HTTP_REQUEST_DEBUG, "http_request_new IN\n");
  req = ymaloc(sizeof(http_request_t), 414);
  req->method = 0;
  req->version = 0;
  req->url_file = 0;
  req->rq_headers = 0;
  req->response_headers = 0;
  req->sock = sock;
  gd_s(HTTP_REQUEST_DEBUG, "http_request_new OUT\n");
  return req;
}

void http_request_delete(http_request_t *req)
{
  gd_s(HTTP_REQUEST_DEBUG, "http_request_delete IN\n");
  if (req) {
    if (req->method) {
      fre_str(&(req->method), 270);
    }
    if (req->version) {
      fre_str(&(req->version), 271);
    }
    if (req->url_file) {
      fre_str(&(req->url_file), 272);
    }

    /* HACK - slightly leaky - we should
     * walk the list and fr.ee the contents first. */
    if (req->rq_headers) {
      gnut_list_fre(req->rq_headers);
    }
    if (req->response_headers) {
      gnut_list_fre(req->response_headers);
    }
    fre_hrt(&req, 273);
  }
  gd_s(HTTP_REQUEST_DEBUG, "http_request_delete OUT\n");
}

/* Break the HTTP request into its parts. */
int http_request_parse(http_request_t *req, char *request)
{
  char *ptr_to_http;
  char *ptr_to_file;
  int len;

  gd_s(HTTP_REQUEST_DEBUG, "http_request_parse IN\n");

  ptr_to_file = strstr(request, " ");
  if (!ptr_to_file) {
    gd_s(HTTP_REQUEST_DEBUG, "http_request_parse OUT - nothing to parse\n");
    return -1;
  }
  ptr_to_file++;

  ptr_to_http=strstr(request,"HTTP/");
  if (ptr_to_http==NULL) {
    gd_s(HTTP_REQUEST_DEBUG, "http_request_parse OUT - not an HTTP/1.0 or higher request\n");
    return -1;
  }

  /*
   * Save the "method" portion of the request.
   */
  len = (ptr_to_file - request);
  req->method = ymaloc(sizeof(char)*(len+1), 415);
  strncpy(req->method, request, len);
  *(req->method+len) = '\0';

  /*
   * Save the "url" portion of the request.
   */
  len = (ptr_to_http-1-ptr_to_file);
  req->url_file = ymaloc(sizeof(char)*(len+1), 416);
  strncpy(req->url_file, ptr_to_file, len);
  *(req->url_file+len) = '\0';

  /*
   * Save the HTTP version of the request - just in case we need it.
   */
  len = strlen(ptr_to_http);
  req->version = ymaloc(sizeof(char)*(len+1), 417);
  strncpy(req->version, ptr_to_http, len);
  *(req->version + len) = '\0';

  gd_s(HTTP_REQUEST_DEBUG, "http_request_parse OUT\n");

  return 0;
}

/* Examine the HTTP headers and find one of the values. */
char *http_request_find_header(http_request_t *req, char *name)
{
  Gnut_List *list;
  http_header_t *hdr;

  gd_s(HTTP_REQUEST_DEBUG, "http_request_find_header IN - ");
  gd_s(HTTP_REQUEST_DEBUG, name);
  gd_s(HTTP_REQUEST_DEBUG, "\n");

  list = req->rq_headers;

  while (1) {
    if (!list) break;
    hdr = list->data;
    if (!hdr) break;

    if (!strcmp(hdr->name, name)) {
      gd_s(HTTP_REQUEST_DEBUG, "http_request_find_header OUT - found header ");
      gd_s(HTTP_REQUEST_DEBUG, hdr->value);
      gd_s(HTTP_REQUEST_DEBUG, "\n");
      return hdr->value;
    }

    list = gnut_list_next(list);
  }

  gd_s(HTTP_REQUEST_DEBUG, "http_request_find_header OUT - no such header\n");

  return 0;
}

/* Read HTTP headers into the request object. */
int http_request_read_header(http_request_t *req)
{
  char buf[4096];
  int ret;
  http_header_t *hdr;

  gd_s(HTTP_REQUEST_DEBUG, "http_request_read_header IN\n");

  /* We must first read the remaining headers
   * and ignore them. */
  ret = 1;
  while (1) {
    errno = 0;
    ret = read_line(req->sock, buf, 4096);
    if (gc_debug_opts & 4) {
      printf("hrh01 %s", buf);
    }
    if (errno == EINTR )
      break;
    if (strlen(buf) < 3)
      break;

    hdr = http_header_new();
    if (!hdr)
      break;

    ret = http_header_parse(hdr, buf);
    if (ret)
      break;

    gd_s(1, "Got header ");
    gd_s(1, hdr->name);
    gd_s(1, ":");
    gd_s(1, hdr->value);
    gd_s(1, "\n");

    req->rq_headers = gnut_list_prepend(req->rq_headers, hdr);
  }
  gd_s(HTTP_REQUEST_DEBUG, "http_request_read_header OUT\n");
  return 0;
}

/* Set the contents of an HTTP header. */
int http_response_header_add(http_request_t *req, char *name, char *value)
{
  http_header_t *hdr;

  gd_s(HTTP_HEADER_DEBUG, "http_response_header_add IN - ");
  gd_s(HTTP_HEADER_DEBUG, name);
  gd_s(HTTP_HEADER_DEBUG, " - ");
  gd_s(HTTP_HEADER_DEBUG, value);
  gd_s(HTTP_HEADER_DEBUG, "\n");

  hdr = http_header_new();
  http_header_set(hdr, name, value);
  req->response_headers = gnut_list_prepend(req->response_headers, hdr);

  gd_s(HTTP_HEADER_DEBUG, "http_response_header_add OUT\n");
  return 0;
}

int http_response_header_add_va(http_request_t *req, char *name, char *format, ...)
{
  va_list argp;
  char hdr_value[64];
  int ret;

  gd_s(HTTP_HEADER_DEBUG, "http_response_header_add_va IN\n");

  va_start(argp,format);
  vsprintf(hdr_value,format,argp);
  va_end(argp);
  ret = http_response_header_add(req, name, hdr_value);

  gd_s(HTTP_HEADER_DEBUG, "http_response_header_add_va OUT\n");
  return ret;
}

http_cookie_t *http_cookie_new(void)
{
  http_cookie_t *cookp;

  cookp = ymaloc(sizeof(http_cookie_t), 418);
  cookp->value = NULL;
  cookp->path = NULL;
  cookp->domain = NULL;
  cookp->expiration = time(NULL);

  return cookp;
}

int http_header_get_cookie(http_request_t *req, http_cookie_t *cookp)
{
  char *end_of_value;
  int len_of_value;
  char *header_value;

  header_value = http_request_find_header(req, "Cookie");
  if (!header_value) return -1;

  end_of_value = strchr(header_value, ';');
  if (!end_of_value) {
    len_of_value = strlen(header_value)+1;
  }
  else {
    len_of_value = end_of_value - header_value;
  }

  cookp->value = ymaloc(sizeof(char)*len_of_value, 419);
  if (!cookp->value) {
    return -1;
  }
  strncpy(cookp->value, header_value, len_of_value);
  cookp->value[len_of_value-1] = '\0';

  return 0;
}

void http_cookie_delete(http_cookie_t *cookie)
{
  if (cookie) {
    if (cookie->value) {
      fre_str(&(cookie->value), 274);
    }
    if (cookie->path) {
      fre_str(&(cookie->path), 275);
    }
    if (cookie->domain) {
      fre_str(&(cookie->domain), 276);
    }
    fre_hct(&cookie, 378);
  }
}

int http_response_set_cookie(http_request_t *req, http_cookie_t *cookie)
{
  return http_response_header_add_va(req, "Set-Cookie", "%s; path=%s", cookie->value, "/");
}

int http_response_header_write(http_request_t *req)
{
  http_header_t *hdr;
  Gnut_List *list;

  gd_s(HTTP_REQUEST_DEBUG, "http_response_header_write IN\n");

  /* If returning a partial file, we would want to write
   * "206 Partial Content", but alas, the original Gnutella
   * (Nullsoft Windows client) doesn't work if we do that */
  http_write_str(req, "HTTP/1.0 200 OK\r\n");

  list = req->response_headers;

  while (1) {
    if (!list) break;
    hdr = list->data;
    if (!hdr) break;
    http_write_va(req, HTTP_MAX_HEADER_SIZE, "%s: %s\r\n", hdr->name, hdr->value);
    list = gnut_list_next(list);
  }

  http_write_str(req, "\r\n");

  gd_s(HTTP_REQUEST_DEBUG, "http_response_header_write OUT\n");

  return 0;
}

int http_write_str(http_request_t *req, char *a)
{
  int ret;

  gd_s(HTTP_IO_DEBUG, "http_write_str IN ");
  gd_s(HTTP_IO_DEBUG, a);
  gd_s(HTTP_IO_DEBUG, "\n");

  if (!req) return -1;
  if (req->sock < 0) return -1;
  ret = write(req->sock, a, strlen(a));

  gd_s(HTTP_IO_DEBUG, "http_write_str OUT ");
  gd_i(HTTP_IO_DEBUG, ret);
  gd_s(HTTP_IO_DEBUG, "\n");
  return ret;

}

int http_write_va(http_request_t *req, int maxlen, char *format, ...)
{
  va_list argp;
  int ret;
  char *value;

  gd_s(HTTP_IO_DEBUG, "http_write_va IN\n");

  value = ymaloc(sizeof(char)*maxlen, 306);

  va_start(argp,format);
  vsprintf(value,format,argp);
  va_end(argp);

  ret = http_write_str(req, value);

  fre_str(&value, 379);

  gd_s(HTTP_IO_DEBUG, "http_write_va OUT ");
  gd_i(HTTP_IO_DEBUG, ret);
  gd_s(HTTP_IO_DEBUG, "\n");

  return ret;
}

/* Create a new header object. */
http_header_t *http_header_new(void)
{
  http_header_t *hdr;

  gd_s(HTTP_HEADER_DEBUG, "http_header_new IN\n");

  hdr = (http_header_t *)ymaloc(sizeof(http_header_t), 307);
  assert(hdr!=NULL);
  hdr->name = NULL;
  hdr->value = NULL;

  gd_s(HTTP_HEADER_DEBUG, "http_header_new OUT\n");

  return hdr;
}

/* Parse an HTTP header from a header line. */
int http_header_parse(http_header_t *hdr, char *hdr_line)
{
  char *value;
  char *s;
  int len;

  gd_s(HTTP_HEADER_DEBUG, "http_header_parse IN\n");

  value = strchr(hdr_line, ':');
  if (!value) {
    gd_s(HTTP_HEADER_DEBUG, "This is not a proper header\n");
    return -1;
  }

  len = value - hdr_line;
  hdr->name = ymaloc(sizeof(char)*(len+1), 308);
  assert(hdr->name!=NULL);

  strncpy(hdr->name, hdr_line, len);
  *(hdr->name+len) = '\0';

  /*
   * Skip past the ":" and the space.
   */
  value++;
  if (*value) value++;

  if (!(*value)) {
    fre_str(&(hdr->name), 380);
    fre_hth(&hdr, 381);
    gd_s(HTTP_HEADER_DEBUG, "http_header_parse OUT - not a proper header\n");
    return -1;
  }

  len = strlen(value)-2;
  hdr->value = ymaloc(sizeof(char)*(len+1), 309);
  assert(hdr->value!=NULL);

  strncpy(hdr->value, value, len);
  *(hdr->value+len) = '\0';

  s = hdr->value+(len-1);
  while (isspace(*s)) *s = '\0';

  gd_s(HTTP_HEADER_DEBUG, "http_header_parse OUT\n");

  return 0;
}

/* Set the contents of an HTTP header. */
int http_header_set(http_header_t *hdr, char *name, char *value)
{
  gd_s(HTTP_HEADER_DEBUG, "http_header_set IN\n");
  hdr->name = ystdup(name, 455);
  hdr->value = ystdup(value, 456);
  gd_s(HTTP_HEADER_DEBUG, "http_header_set OUT\n");
  return 0;
}

void http_header_delete(http_header_t *hdr)
{
  gd_s(HTTP_HEADER_DEBUG, "http_header_delete IN\n");
  if (hdr) {
    if (hdr->name) {
      fre_str(&(hdr->name), 376);
    }
    if (hdr->value) {
      fre_str(&(hdr->value), 377);
    }
    fre_hth(&hdr, 378);
  }
  gd_s(HTTP_HEADER_DEBUG, "http_header_delete OUT\n");
}
